#ifndef EXERCISE_3_ERRORESTIMATOR_H_
#define EXERCISE_3_ERRORESTIMATOR_H_

/** Error estimator for Laplace's equation
 *
 *  \file errorEstimator.h
 */

#include <typeinfo>
#include <cxxabi.h>

#include <Eigen/Dense>

/**
 *  \brief Class for local error estimation
 */
template<typename ConfiguratorType>
class ErrorEstimator
{
public:
    typedef typename ConfiguratorType::InitType       MeshType;
    typedef typename ConfiguratorType::RealType       RealType;
    typedef typename ConfiguratorType::DomVecType     DomVecType;
    typedef typename ConfiguratorType::TangentVecType TangentVecType;
    // Strangely, lpNorm fails if this is set to
    // "typename ConfiguratorType::VectorType". Bug?
    typedef typename Eigen::VectorXd VectorType;
    typedef typename Eigen::Matrix<RealType, 3, 1> VecType;

    //! \brief Constructor
    //! \param[in] mesh The finite element mesh
    //! \param[in] config The configurator used with the mesh
    //! \param[in] numericalSolution The numerical solution to be used in the error estimate
    //! \param[in] gamma Parameter determining elements containing which
    //!                  fraction of the total error will be refined (between 0 and 1)
    ErrorEstimator(MeshType &mesh, const ConfiguratorType &config,
            const VectorType &numericalSolution,
            RealType gamma = 0.4)
    : _mesh(mesh),
      _numericalSolution(config, numericalSolution), _gamma(gamma)
    {}

    RealType markAndRefineTriangles(){
        RealType err = markTriangles();
        _mesh.refineMarkedTriangles();
        return err;
    }

    //! \brief Mark triangles based on the error estimates
    RealType markTriangles()
    {
        VectorType errEstimates(_mesh.getNumTriangs());

        computeLocalErrorEstimates(errEstimates);

        //! \todo Mark triangles using errEstimates

        return errEstimates.norm();
    }

    //! \brief Compute error estimator for each element
    void computeLocalErrorEstimates(VectorType &errEstimates) const
    {
        //! \todo Fill errEstimates with error estimates per element
    }

    //! \brief Compute length on longest edge for each element
    void computeH(VectorType &h) const
    {
        for (int t = 0; t < _mesh.getNumTriangs(); ++t) {
            h[t] = _mesh.getTriang(t).getEdge(_mesh.getLongestEdgeIndex(t)).norm();
        }
    }

    //! \brief Returns reference to the mesh that was used for initialization of this object
    MeshType& getMesh()
    {
        return _mesh;
    }

protected:

    MeshType &_mesh;
    const shellFE::DiscreteFunctionDefaultShellFE<ConfiguratorType> _numericalSolution;
    RealType _gamma;
};

#endif /* EXERCISE_3_ERRORESTIMATOR_H_ */
